# -*- encoding: utf-8 -*-
##############################################################################
#
#    Acrisel LTD
#    Copyright (C) 2008- Acrisel (acrisel.com) . All Rights Reserved
#
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see http://www.gnu.org/licenses/.
#
##############################################################################
import logging
from logging.handlers import QueueListener, QueueHandler
import os
import multiprocessing as mp
from acris.timed_sized_logging_handler import TimedSizedRotatingHandler

class MpQueueListener(QueueListener):
    def __init__(self, queue, *handlers):
        super(MpQueueListener, self).__init__(queue, *handlers)
        """
        Initialise an instance with the specified queue and
        handlers.
        """
        # Changing this to a list from tuple in the parent class
        self.handlers = list(handlers)

    def handle(self, record):
        """
        Override handle a record.

        This just loops through the handlers offering them the record
        to handle.

        :param record: The record to handle.
        """
        record = self.prepare(record)
        for handler in self.handlers:
            if record.levelno >= handler.level: # This check is not in the parent class
                handler.handle(record)

    def addHandler(self, hdlr):
        """
        Add the specified handler to this logger.
        """
        if not (hdlr in self.handlers):
            self.handlers.append(hdlr)

    def removeHandler(self, hdlr):
        """
        Remove the specified handler from this logger.
        """
        if hdlr in self.handlers:
            hdlr.close()
            self.handlers.remove(hdlr)

class MpLogger(object):
    logger_initialized=False
    queue_listener=None
    
    def __init__(self, logdir=None, logging_level=logging.INFO, record_format='', logging_root=None):
        self.logdir=logdir
        self.logging_level=logging_level
        self.record_format=record_format
        self.logging_root=logging_root

    def start(self, ):
        ''' starts logger for multiprocessing using queue.
        
        logdir: if provided, error and debug logs will be created in it.
        logging_level
        logger_format
        '''
        # create console handler and set level to info
        #global logger_initialized
        #global queue_listener
        
        if MpLogger.logger_initialized:
            return
        
        if not self.record_format:
            record_format="[ %(asctime)s ][ %(levelname)s ][ %(message)s ][ %(module)s.%(funcName)s ]"
    
        MpLogger.logger_initialized=True
        logger = logging.getLogger(name=self.logging_root)
        logger.setLevel(self.logging_level)
            
        q=mp.Queue()
        queue_handler = QueueHandler(q)
        logger.addHandler(queue_handler)
        MpLogger.queue_listener = MpQueueListener(q,)
    
        handler = logging.StreamHandler()
        handler.setLevel(self.logging_level)
        formatter = logging.Formatter(record_format)
        handler.setFormatter(formatter)
        #logger.addHandler(handler)
        MpLogger.queue_listener.addHandler(handler)
    
        if self.logdir:
            # create error file handler and set level to error
            handler = TimedSizedRotatingHandler(filename=os.path.join(self.logdir, "error.log"), encoding=None, delay="true")
            handler.setLevel(logging.ERROR)
            #formatter = logging.Formatter("%(levelname)s - %(message)s")
            formatter = logging.Formatter(record_format)
            handler.setFormatter(formatter)
            #logger.addHandler(handler)
            MpLogger.queue_listener.addHandler(handler)
         
            # create debug file handler and set level to debug
            handler = TimedSizedRotatingHandler(filename=os.path.join(self.logdir, "debug.log"), )
            handler.setLevel(logging.DEBUG)
            formatter = logging.Formatter(record_format)
            handler.setFormatter(formatter)
            #logger.addHandler(handler)
            MpLogger.queue_listener.addHandler(handler)
            
        MpLogger.queue_listener.start()
        
    def stop(self,):
        #global queue_listener
        if MpLogger.queue_listener:
            MpLogger.queue_listener.stop()
            
    def quite(self,):
        if MpLogger.queue_listener:
            MpLogger.queue_listener.enqueue_sentinel()

 
if __name__ == '__main__':
    import time
    import random
    
    logger=logging.getLogger(__name__)
    
    def subproc(limit=1):
        for i in range(limit):
            sleep_time=3/random.randint(1,10)
            time.sleep(sleep_time)
            logger.info("proc [%s]: %s/%s - sleep %4.4ssec" % (os.getpid(), i, limit, sleep_time))
    
    mplogger=MpLogger(logging_level=logging.DEBUG)
    mplogger.start()
    
    logger.debug("starting sub processes")
    procs=list()
    for limit in [5, 5]:
        proc=mp.Process(target=subproc, args=(limit, ))
        procs.append(proc)
        proc.start()
        
    for proc in procs:
        if proc:
            proc.join()
        
    logger.debug("sub processes completed")

    mplogger.stop()


__version__ = "1.1.3"

from .threaded import threaded
from .singleton import Singleton
from .sequence import Sequence
from .timed_sized_logging_handler import TimedSizedRotatingHandler
from .mplogger import MpLogger



'''
Created on Nov 23, 2016

@author: arnon
'''
# -*- encoding: utf-8 -*-
##############################################################################
#
#    Acrisel LTD
#    Copyright (C) 2008- Acrisel (acrisel.com) . All Rights Reserved
#
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see http://www.gnu.org/licenses/.
#
##############################################################################

from logging.handlers import RotatingFileHandler, TimedRotatingFileHandler

class TimedSizedRotatingHandler(TimedRotatingFileHandler, RotatingFileHandler):
    
    def __init__(self, filename, mode='a', maxBytes=0, backupCount=0, encoding=None, delay=False, when='h', interval=1, utc=False, atTime=None):
        """ Combines RotatingFileHandler TimedRotatingFileHandler)  """
        RotatingFileHandler.__init__(self, filename=filename, mode=mode, maxBytes=maxBytes, backupCount=backupCount, encoding=encoding, delay=delay)
        TimedRotatingFileHandler.__init__(self, filename=filename, when=when, interval=interval, backupCount=backupCount, encoding=encoding, delay=delay, utc=utc, atTime=atTime)

    def shouldRollover(self, record):
        """
        Check the need to rotate.     
        """
        timed_rollover=TimedRotatingFileHandler.shouldRollover(self, record) 
        sized_rollover=RotatingFileHandler.shouldRollover(self, record)
        
        return timed_rollover or sized_rollover

    def doRollover(self):
        """
        It is enough to use timed base rollover.
        """
        super(TimedRotatingFileHandler, self).doRollover()

    def getFilesToDelete(self):
        """
        It is enough to use timed base rollover.
        """
        return super(TimedRotatingFileHandler, self).getFilesToDelete()            


# -*- encoding: utf-8 -*-
##############################################################################
#
#    Acrisel LTD
#    Copyright (C) 2008- Acrisel (acrisel.com) . All Rights Reserved
#
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see http://www.gnu.org/licenses/.
#
##############################################################################


from threading import Lock


class SingletonType(type):
    __locker=Lock() 
    #__instance = None
    __instance={}
      
    def __call__( self, name='', *args, **kargs):   
        SingletonType.__locker.acquire()
        try:
            instance=self.__instance[name]
        except KeyError:
            instance=super(SingletonType, self).__call__(*args, **kargs)
            self.__instance[name]=instance
        #if not self.__instance:
        #    self.__instance = super(SingletonType, self).__call__(*args, **kargs)
        SingletonType.__locker.release()
        return instance
    
        
class Singleton(metaclass=SingletonType):
    pass


# -*- encoding: utf-8 -*-
##############################################################################
#
#    Acrisel LTD
#    Copyright (C) 2008- Acrisel (acrisel.com) . All Rights Reserved
#
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see http://www.gnu.org/licenses/.
#
##############################################################################

"""
About
=========
:synopsis:     threaded provides decorator for class methods to run via a thread
:moduleauthor: Arnon Sela
:date:         2016/10/28
:description:  a method decorated with @threaded would run in a separated thread. this work 
   
:comment: based on http://code.activestate.com/recipes/576684-simple-threading-decorator/


      
**History:**
-------------------

:Author: <who made the changes>
:Modification:
   - Describe here the changes that were made in bullet form.  Use
     a new dash character '-' to create a new entry!
:Date: <enter the date the changes were made>

    
"""


from functools import wraps
from threading import Thread, RLock, Event #, current_thread

__all__ = ['threaded']

class AsyncResult(object): 
    """Represents an asynchronous operation that may not have completed yet.""" 
    def __init__(self): 
        self.completed = False 
        self.failed = False 
        self.__wait = Event() 
        self.__callbacks = [] 
        self.__errbacks = [] 
        self.__retval = None 
        self.__error = None 
        self.__lock = RLock()

    def complete(self):
        self.__lock.acquire()
        self.completed = True
        self.__wait.set()
        self.__lock.release()
    
    def succeed(self, retval):
        self.__retval = retval
        self.complete()
        for callback in self.__callbacks:
            callback(retval)
        self.clearCallbacks()
    
    def fail(self, error):
        self.__error = error
        self.failed = True
        self.complete()
        for errback in self.__errbacks:
            errback(error)
        self.clearCallbacks()
    
    def clearCallbacks(self):
        self.__callbacks = []
        self.__errbacks = []
    
    def addCallback(self, callback, errback=None):
        self.__lock.acquire()
        try:
            if self.completed:
                if not self.failed:
                    callback(self.__retval)
            else:
                self.__callbacks.append(callback)
            if not errback == None:
                self.addErrback(errback)
        finally:
            self.__lock.release()
    
    def addErrback(self, errback):
        self.__lock.acquire()
        try:
            if self.completed:
                if self.failed:
                    errback(self.__error)
            else:
                self.__errbacks.append(errback)
        finally:
            self.__lock.release()
    
    def __getResult(self):
        self.__wait.wait()
        if not self.failed:
            return self.__retval
        else:
            raise self.__error
    result=property(__getResult)
    
def threaded(method): 
    @wraps(method)
    def wrapper(*args, **kwargs): 
        async_result = AsyncResult() 
        def _method(): 
            try: 
                result=method(*args, **kwargs) 
            except Exception as e: 
                print(e)
                async_result.fail(e)
            else:
                async_result.succeed(result)
        Thread(target = _method).start() 
        return async_result 
    return wrapper